"""Authentication backends for the API Gateway."""

import contextlib
import logging

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import RemoteUserBackend

from mitol.apigateway.api import decode_x_header

log = logging.getLogger(__name__)
User = get_user_model()


class RemoteUserCustomFieldBackend(RemoteUserBackend):
    """
    RemoteUserBackend variant that allows the field for the lookup to be configured
    """

    lookup_field: str

    def authenticate(self, request, remote_user):
        """
        Authenticate the user
        """
        if not remote_user:
            return None
        created = False
        user = None
        username = self.clean_username(remote_user)

        if self.create_unknown_user:
            user, created = User.objects.get_or_create(**{self.lookup_field: username})
        else:
            with contextlib.suppress(User.DoesNotExist):
                user = User.objects.get_by_natural_key(username)
        user = self.configure_user(request, user, created=created)
        return user if self.user_can_authenticate(user) else None

    async def aauthenticate(self, request, remote_user):
        """See authenticate()."""
        if not remote_user:
            return None
        created = False
        user = None
        username = self.clean_username(remote_user)

        if self.create_unknown_user:
            user, created = await User.objects.aget_or_create(
                **{self.lookup_field: username}
            )
        else:
            with contextlib.suppress(User.DoesNotExist):
                user = await User.objects.aget_by_natural_key(username)
        user = await self.aconfigure_user(request, user, created=created)
        return user if self.user_can_authenticate(user) else None


class ApisixRemoteUserBackend(RemoteUserCustomFieldBackend):
    """
    Custom RemoteUserBackend that updates users using the APISIX headers.

    RemoteUserBackend already has some support for creating an unknown user, but
    it won't fill out all the data we'll generally want to capture. Additionally,
    we'll want to toggle the user creation code with a setting.
    """

    lookup_field = "global_id"

    create_unknown_user = settings.MITOL_APIGATEWAY_USERINFO_CREATE
    update_known_user = settings.MITOL_APIGATEWAY_USERINFO_UPDATE

    def configure_user(self, request, user, *, created=True):
        """
        Configure the user - use the mapping to fill out the object(s).

        See MITOL_APIGATEWAY_USERINFO_MODEL_MAP in settings.py for the mapping.
        See also the flags above to configure when this updates the user object.
        """

        if not created and not self.update_known_user:
            log.debug("configure_user: Not updating known user %s", user)
            return user

        if created and not self.create_unknown_user:
            log.debug("configure_user: Not updating created user %s", user)
            return user

        infomap = settings.MITOL_APIGATEWAY_USERINFO_MODEL_MAP
        decoded_headers = decode_x_header(request)

        for header_field, model_field in infomap["user_fields"].items():
            setattr(user, model_field, decoded_headers.get(header_field, None))

        user.save()

        log.debug("configure_user: Updated user %s", user)

        for model_name in infomap["additional_models"]:
            AdditionalModel = apps.get_model(model_name)
            model_fields = {
                "user": user,
            }

            for header_field, model_field, default_value in infomap[
                "additional_models"
            ][model_name]:
                model_fields[model_field] = decoded_headers.get(
                    header_field, default_value
                )

            addl_model, _ = AdditionalModel.objects.update_or_create(
                user=user,
                defaults=model_fields,
            )

            log.debug("configure_user: Updated model %s: %s", model_name, addl_model)

        return user



"""API functions."""

import base64
import json
import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import HttpRequest

log = logging.getLogger(__name__)
User = get_user_model()


def decode_x_header(request: HttpRequest | dict) -> dict | None:
    """
    Decode an 'X-' header.

    API gateways put the user information in a header, typically X-Userinfo, as
    a base64 encoded JSON string. This function decodes and de-JSONs that string
    and returns the resulting dict.

    This will work for both HttpRequest and Channel scopes.

    Args:
        request (HttpRequest): the HTTP request
    Returns:
    dict of decoded values, or None if the header isn't found
    """

    if isinstance(request, HttpRequest):
        x_userinfo = request.META.get(
            settings.MITOL_APIGATEWAY_USERINFO_HEADER_NAME, False
        )
    else:
        # We've likely got a scope dict from Channels.
        # We need to do some processing - scopes don't append HTTP_, don't upper
        # the header name, and don't convert - to _. (Also these get stored as
        # bytes.)
        check_header = (
            settings.MITOL_APIGATEWAY_USERINFO_HEADER_NAME.lower()
            .replace("http_", "")
            .replace("_", "-")
        )
        scope_headers = request.get("headers", [])
        x_userinfo = [x[1] for x in scope_headers if x[0].decode() == check_header]

        if len(x_userinfo) == 0:
            log.warning("No %s header found", check_header)
            return None

        x_userinfo = x_userinfo.pop()

    if not x_userinfo:
        log.warning(
            "No %s header found", settings.MITOL_APIGATEWAY_USERINFO_HEADER_NAME
        )
        return None

    decoded_x_userinfo = base64.b64decode(x_userinfo)
    return json.loads(decoded_x_userinfo)


def get_user_id_from_userinfo_header(request: HttpRequest | dict) -> str | None:
    """Return the user ID from the userinfo header."""

    decoded_headers = decode_x_header(request)

    if not decoded_headers:
        return None

    return decoded_headers.get(settings.MITOL_APIGATEWAY_USERINFO_ID_FIELD, None)


def get_username_from_userinfo_header(request: HttpRequest | dict) -> str | None:
    """Return the username from the userinfo header."""

    user_id = get_user_id_from_userinfo_header(request)

    if not user_id:
        return None

    return User.objects.get(global_id=user_id).username


def create_userinfo_header(user):
    """
    Create an X_USERINFO header, so we can fake out auth properly.
    APIClient has a force_authenticate method that adds the user account to the
    fake request, but the middleware in apigateway will kick it out because the
    header won't be there. So, use this to add the header when creating APIClients.
    Ex: new APIClient(headers=create_userinfo_header(myuser))
    This will use the model map in reverse. It will only care about the
    user_fields
    """

    model_map = settings.MITOL_APIGATEWAY_USERINFO_MODEL_MAP
    header_name = settings.MITOL_APIGATEWAY_USERINFO_HEADER_NAME.replace("HTTP_", "")
    header_data = {
        k: str(getattr(user, model_map["user_fields"][k], None))
        for k in model_map["user_fields"]
    }

    return {header_name: base64.b64encode(json.dumps(header_data).encode())}



"""Custom logout view for the API Gateway."""

import logging

from django.conf import settings
from django.contrib.auth import logout
from django.shortcuts import redirect
from django.utils.http import url_has_allowed_host_and_scheme
from django.views import View

log = logging.getLogger(__name__)


def get_redirect_url(request):
    """
    Get the redirect URL from the request.

    Args:
        request: Django request object

    Returns:
        str: Redirect URL
    """
    log.debug("views.get_redirect_url: Request GET is: %s", request.GET.get("next"))
    log.debug(
        "views.get_redirect_url: Request cookie is: %s", request.COOKIES.get("next")
    )

    next_url = request.GET.get("next") or request.COOKIES.get("next")
    log.debug("views.get_redirect_url: Redirect URL (before valid check): %s", next_url)

    if request.COOKIES.get("next"):
        # Clear the cookie after using it
        log.debug("views.get_redirect_url: Popping the next cookie")

        request.COOKIES.pop("next", None)

    return (
        next_url
        if next_url
        and url_has_allowed_host_and_scheme(
            next_url, allowed_hosts=settings.MITOL_APIGATEWAY_ALLOWED_REDIRECT_HOSTS
        )
        else settings.MITOL_APIGATEWAY_DEFAULT_POST_LOGOUT_DEST
    )


class ApiGatewayLogoutView(View):
    """
    Log the user out.

    Logs the user out of the Django app itself. If there is also an APISIX session,
    send them through APISIX's logout process too (which will destroy the APISIX
    session and log them out in SSO/Keycloak). If there's a redirect URL set
    (in cookies or in the query), send the user there; otherwise, use the default
    logout URL.

    If we just send the user through the APISIX logout, they'll get an error
    message if they don't have an APISIX session. (They'll still get sent to
    Keycloak to log out, but they won't have all the necessary data to that, and
    Keycloak will throw an error.)
    """

    def get(
        self,
        request,
        *args,  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ):
        """
        GET endpoint reached after logging a user out from Keycloak
        """
        user = getattr(request, "user", None)
        user_redirect_url = get_redirect_url(request)
        log.debug(
            "views.ApiGatewayLogoutView.get: User redirect URL: %s", user_redirect_url
        )
        if user and user.is_authenticated:
            logout(request)

        if request.META.get(settings.MITOL_APIGATEWAY_USERINFO_HEADER_NAME):
            # Still logged in via Apisix/Keycloak, so log out there as well
            log.debug("views.ApiGatewayLogoutView.get: Send to APISIX logout URL")
            return redirect(settings.MITOL_APIGATEWAY_LOGOUT_URL)
        else:
            log.debug("views.ApiGatewayLogoutView.get: Send to %s", user_redirect_url)
            return redirect(user_redirect_url)



"""
Middleware to fetch the user out of the headers.
Middleware for channels is in middleware_channels.py.
"""

import logging

from django.conf import settings
from django.contrib.auth.middleware import (
    PersistentRemoteUserMiddleware,
    RemoteUserMiddleware,
)

from mitol.apigateway.api import get_user_id_from_userinfo_header

log = logging.getLogger(__name__)


class ApisixUserMiddleware(RemoteUserMiddleware):
    """Checks for and processes APISIX-specific headers."""

    def process_request(self, request):
        """
        Modify the header to contaiin username, pass off to RemoteUserMiddleware
        """

        log.debug("ApisixUserMiddleware.process_request: started")

        if settings.MITOL_APIGATEWAY_DISABLE_MIDDLEWARE:
            return self.get_response(request)

        if request.META.get(settings.MITOL_APIGATEWAY_USERINFO_HEADER_NAME):
            new_header = get_user_id_from_userinfo_header(request)
            request.META["REMOTE_USER"] = new_header

        super().process_request(request)

        response = self.get_response(request)

        next_param = request.GET.get("next", None) if request.GET else None
        if next_param:
            log.debug(
                "ApisixUserMiddleware.process_request: Setting next cookie to %s",
                next_param,
            )
            response.set_cookie("next", next_param, max_age=30, secure=False)

        log.debug(
            "ApisixUserMiddleware.process_request: Next cookie is %s",
            response.cookies.get("next"),
        )

        return response


class PersistentApisixUserMiddleware(
    PersistentRemoteUserMiddleware, ApisixUserMiddleware
):
    """Persistent version of the ApisixUserMiddleware."""



"""
Middleware for Django Channels.

The RemoteUserMiddleware doesn't put the user information into the session, so
the AuthMiddleware that ships with Channels doesn't work. So, this middleware
(and related functions) work more like the RemoteUserMiddleware - we attach the
user to the scope but it's not in the session. There is no real need to login.

Important consideration for websocket consumers: the headers get attached when
the connection is made. This means that user change events don't propogate into
_running_ consumers. You will need to check and perform appropriate actions when
the user changes through some other means, or your websocket users will continue
to be logged in even if they've logged out elsewhere.
"""

import logging

from channels.auth import AuthMiddleware
from channels.db import database_sync_to_async
from channels.sessions import CookieMiddleware, SessionMiddleware

from mitol.apigateway.api import get_user_id_from_userinfo_header
from mitol.apigateway.backends import ApisixRemoteUserBackend

log = logging.getLogger(__name__)


@database_sync_to_async
def get_apisix_user(scope):
    """
    Get the user using the ApisixRemoteUserBackend.
    """

    user_id = get_user_id_from_userinfo_header(scope)

    log.debug("Got user ID %s", user_id)

    if user_id:
        backend = ApisixRemoteUserBackend()
        return backend.authenticate(request=scope, remote_user=user_id)

    from django.contrib.auth.models import AnonymousUser

    return AnonymousUser()


class ApisixUserMiddleware(AuthMiddleware):
    """Retrieve the current user based on the APISIX headers provided."""

    async def resolve_scope(self, scope):
        """Resolve the user from the scope."""

        scope["user"]._wrapped = await get_apisix_user(scope)  # noqa: SLF001


def ApisixAuthMiddlewareStack(inner):  # noqa: N802
    """Wrap the ApisixUserMiddleware in a couple of other useful middlewares."""

    return CookieMiddleware(SessionMiddleware(ApisixUserMiddleware(inner)))
